<?php

namespace Drupal\entity_update;

use Drupal\Component\Serialization\Json;
use Drupal\Core\Database\Database;

/**
 * EntityUpdate Main Class.
 */
class EntityUpdate {

  /**
   * Update all empty entities.
   */
  public static function basicUpdate() {
    // Check all updateble entities are empty.
    $list = self::getEntityTypesToUpdate();
    $flgOK = TRUE;
    foreach ($list as $item => $entity_type_changes) {
      if (!empty(\Drupal::entityQuery($item)->execute())) {
        $flgOK = FALSE;
        drush_log("The entity '$item' is not empty", 'warning');
      }
    }

    // Return if one of the entity has data.
    if (!$flgOK) {
      drush_log("Cant update ad basic, use --all", 'error');
      return FALSE;
    }

    // Update.
    try {
      \Drupal::entityDefinitionUpdateManager()->applyUpdates();
      return TRUE;
    }
    catch (\Exception $e) {
      drush_log($e->getMessage(), 'warning');
    }
    return FALSE;
  }

  /**
   * Update all entities.
   *
   * @return bool
   *   Update is done or FAIL
   */
  public static function safeUpdateMain() {

    // Get entitiy types to update.
    $entity_change_summerys = self::getEntityTypesToUpdate();
    if (empty($entity_change_summerys)) {
      drush_log('No entities to update.', 'cancel');
      return FALSE;
    }

    // Select the method to use.
    $flg_has_install = FALSE;
    $flg_has_uninstall = FALSE;
    foreach ($entity_change_summerys as $entity_type_changes) {
      foreach ($entity_type_changes as $entity_change_summ) {

        if (strstr($entity_change_summ, "uninstalled")) {
          $flg_has_uninstall = TRUE;
        }
        else {
          $flg_has_install = TRUE;
        }
      }
    }

    // Install Or Uninstall not found.
    if (!$flg_has_install && !$flg_has_uninstall) {
      drush_log('Entity install / Uninstall not found.', 'cancel');
      return FALSE;
    }
    elseif ($flg_has_install && $flg_has_uninstall) {
      drush_log('Has fields to install and to uninstall. Do one action at a time.', 'cancel');
      return FALSE;
    }
    elseif ($flg_has_install) {
      return self::safeUpdateInstallFields();
    }
    elseif ($flg_has_uninstall) {
      return self::safeUpdateUninstallFields($entity_change_summerys);
    }

    drush_log('UNKNOWN ERROR.', 'error');
    return FALSE;
  }

  /**
   * Update all entities / Uninstall fields.
   *
   * @return bool
   *   Update is done or FAIL
   */
  private static function safeUpdateUninstallFields($entity_change_summerys) {

    // Get Database connection.
    $con = Database::getConnection();

    // Flag entities has data.
    $flg_has_data = FALSE;
    // Read and backup data into entity_update entity..
    drush_print("Read and backup data");
    foreach ($entity_change_summerys as $entity_type_id => $entity_type_changes) {
      drush_print(" - Reading : $entity_type_id");
      $entities = \Drupal::entityTypeManager()
        ->getStorage($entity_type_id)
        ->loadMultiple();

      foreach ($entities as $entity_id => $entity) {
        $entity_type = $entity->getEntityType();

        // Backup entity to DB.
        $con->insert('entity_update')
          ->fields([
            'entity_type' => $entity_type_id,
            'entity_id' => $entity_id,
            'entity_class' => $entity_type->getClass(),
            'status' => 0,
            'data' => Json::encode($entity->toArray()),
          ])
          ->execute();

        // Delete Entity.
        $entity->delete();
        $flg_has_data = TRUE;
      }
    }

    // Exec Update.
    drush_print("Update entity Schema");
    try {
      \Drupal::entityDefinitionUpdateManager()->applyUpdates();
      drush_log('Entities update success', 'ok');
    }
    catch (Exception $ex) {
      drush_log($ex->getMessage(), 'warning');
    }

    // Re create.
    if ($flg_has_data) {
      drush_print("Re creating entities.");
      $db_data = $con->select('entity_update', 't')
        ->fields('t')
        ->execute()
        ->fetchAll(\PDO::FETCH_ASSOC);

      if (!$db_data) {
        drush_log("ERROR, Data read error", 'error');
        return FALSE;
      }

      foreach ($db_data as $row) {

        $entity_class = $row['entity_class'];
        $entity_data = Json::decode($row['data']);

        // Create Entity.
        $entity = $entity_class::create($entity_data);

        // Save entity.
        $entity->save();
      }

      // Message to Flush 'entity_update' via drush command.
      drush_log("Entiti recreate Success / End", 'ok');
      drush_log("CAUTION : Before next operation, Flush 'Entity Data' using : drush upe --clean", 'warning');
    }

    return TRUE;
  }

  /**
   * Update all entities / Install fields.
   *
   * @return bool
   *   Update is done or FAIL
   */
  private static function safeUpdateInstallFields() {
    // Update.
    try {
      \Drupal::entityDefinitionUpdateManager()->applyUpdates();
      drush_log('Entities update success', 'ok');
      return TRUE;
    }
    catch (\Exception $e) {
      drush_log($e->getMessage(), 'warning');
    }
    return FALSE;
  }

  /**
   * Update an entities.
   */
  public static function safeUpdateEntity($entity_type, $check = TRUE) {

    if ($check) {
      $list = self::getEntityTypesToUpdate();
      if (!isset($list[$entity_type])) {
        drush_log("No updates for the entity type '$entity_type'", 'warning');
        return FALSE;
      }
    }

    return 0;
  }

  /**
   * Cleanup entity backup database.
   *
   * @return array
   *   The list of entities to update.
   */
  public static function cleanupEntityBackup() {

    // Get Database connection.
    $con = Database::getConnection();
    $con->truncate('entity_update')->execute();

    return TRUE;
  }

  /**
   * Get Entity types to update.
   *
   * @return array
   *   The list of entities to update.
   */
  public static function getEntityTypesToUpdate() {

    $list = entity_update_get_entity_changes();

    foreach ($list as $entity_type_id => $entity_change_summery) {
      $list[$entity_type_id] = $entity_change_summery;
    }

    return $list;
  }

}
